# -*- coding: utf-8 -*-
"""
Parser and core Routines
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import ast
import re
import sys
from warnings import warn

from six.moves import map, range

unicode_tbl = {
    "α": "alpha",
    "β": "beta",
    "χ": "chi",
    "δ": "delta",
    "ε": "epsilon",
    "γ": "gamma",
    "ψ": "psi",
    "θ": "theta",
    "κ": "kappa",
    "λ": "lambda",
    "η": "eta",
    "ν": "nu",
    "π": "pi",
    "ϕ": "phi",
    "σ": "sigma",
    "τ": "tau",
    "ω": "omega",
    "ξ": "xi",
    "Δ": "Delta",
    "φ": "Phi",
    #    'Φ':'Phi',
    "Γ": "Gamma",
    "Ψ": "Psi",
    "Λ": "Lambda",
    "Σ": "Sigma",
    "Ξ": "Xi",
}

fracs = {
    0.5: ["", 1, 2],
    -0.5: ["-", 1, 2],
    0.75: ["", 3, 4],
    -0.75: ["-", 3, 4],
    0.25: ["", 1, 4],
    -0.25: ["-", 1, 4],
}

# Modules removed from expressions:
clear_modules = [
    "math",
    "np",
    "numpy",
    # Note that scipy.integrate must be placed before scipy as
    # names are removed in this order
    "scipy.integrate",
    "scipy",
    "df",  # not a module, but useful to clear pandas dataframe for readability
]

# % Printing & encoding


def uprint(*expr):
    """ Deals with encoding problems """

    try:
        print(*expr)
    except UnicodeEncodeError:
        f = lambda expr: expr.encode(sys.stdout.encoding, errors="replace")
        print(*list(map(f, expr)))


class LatexVisitor(ast.NodeVisitor):
    """
    Parameters
    ----------

    simplify_multipliers: bool
        if ``True``, simplify expression if multiplier is a float. Ex::

            2*a -> 2a
            a*3.5 -> 3.5a

        see :meth:`~pytexit.core.core.LatexVisitor.visit_BinOp` for more
        information. Default ``True``.

    simplify_fractions: bool
        if ``True``, simplify fractions.
        see :meth:`~pytexit.core.core.LatexVisitor.visit_BinOp` for more
        information. Default ``False``.

    simplify_ints: bool
        see :meth:`~pytexit.core.core.LatexVisitor.visit_BinOp` for more
        information. Default ``False``.

    """
    def __init__(
        self,
        dummy_var,
        upperscript,
        lowerscript,
        verbose,
        simplify_multipliers,
        simplify_fractions,
        simplify_ints,
        tex_multiplier,
    ):

        super(LatexVisitor, self).__init__()
        # super().__init__()  # doesn't work in Python 2.x
        self.dummy_var = dummy_var

        self.upper = upperscript
        self.lower = lowerscript

        self.verbose = verbose
        self.simplify_multipliers = simplify_multipliers
        self.simplify_fractions = simplify_fractions
        self.simplify_ints = simplify_ints
        self.tex_multiplier = tex_multiplier

        self.precdic = {
            "Pow": 700,
            "Div": 400,
            "FloorDiv": 400,
            "Mult": 400,
            "Invert": 800,
            "Compare": 300,
            "UAdd": 800,
            "Not": 800,
            "USub": 800,
            "Num": 1000,
            "Constant": 1000,
            "Assign": 300,
            "Sub": 300,
            "Add": 300,
            "Mod": 500,
            "ListComp": 1000,
            "list": 1000,
            "Call": 1000,
            "Name": 1000,
        }

    def looks_like_int(self, a):
        """ Check if the input ``a`` looks like an integer """

        if self.simplify_ints:
            try:
                return float(str(a).split(".")[1]) == 0.0
            except (IndexError, ValueError):
                return False
        else:
            return False

    def prec(self, n):
        if n.__class__.__name__ in self.precdic:
            return self.precdic[n.__class__.__name__]
        else:
            return getattr(self, "prec_" + n.__class__.__name__,
                           getattr(self, "generic_prec"))(n)

    def visit_ListComp(self, n, kwout=False):
        """Analyse a list comprehension
        Output :
        - kw : used by some functions to display nice formulas (e.g : sum)
        - args : standard output to display a range in other cases
        """

        kw = {}

        #        lc = n.args[0]
        comp = n.generators[0]
        kw["iterator"] = self.visit(comp.target)
        f = self.visit(comp.iter.func)

        if f == "range":
            if len(comp.iter.args) > 1:
                kw["min"] = self.visit(comp.iter.args[0])
                kw["max"] = self.visit(comp.iter.args[1])
            else:
                kw["min"] = 0
                kw["max"] = self.visit(comp.iter.args[0])
            # Remove 1 for range max
            try:
                kw["max"] = int(kw["max"]) - 1
            except ValueError:
                if kw["max"].endswith(r"+1"):
                    # write 'sum([... range(N+1)])' as (sum^N)
                    kw["max"] = kw["max"][:-2]
                else:
                    # write 'sum([... range(N)])' as (sum^N-1)
                    kw["max"] = r"{0}-1".format(kw["max"])
            kw["content"] = self.visit(n.elt)

        args = r"%s, %s=%s..%s" % (kw["content"], kw["iterator"], kw["min"],
                                   kw["max"])

        if kwout:
            return args, kw
        else:
            return args

    def visit_list(self, n):
        self.generic_visit(n)

    def visit_Call(self, n):
        """ Node details : n.args, n.func, n.keywords, n.kwargs"""
        func = self.visit(n.func)

        # Deal with list comprehension and complex formats
        if len(n.args) > 0:
            blist = isinstance(n.args[0], ast.ListComp)
        else:
            blist = False

        if blist:
            args, kwargs = self.visit_ListComp(n.args[0], kwout=True)
        else:
            args = ", ".join(map(self.visit, n.args))

        # Usual math functions
        if func in ["cos", "sin", "tan", "cosh", "sinh", "tanh"]:
            return "{0}{1}".format(func, self.parenthesis(args))
        elif func == "sqrt":
            return self.sqrt(args)
        # by default log refers to log10 in Python. Unless people import it as
        # ln
        elif func in ["log", "ln"]:
            return r"\ln%s" % self.parenthesis(args)
        elif func in ["log10"]:
            return r"\log%s" % self.parenthesis(args)
        elif func in ["arccos", "acos"]:
            return r"\arccos%s" % self.parenthesis(args)
        elif func in ["arcsin", "asin"]:
            return r"\arcsin%s" % self.parenthesis(args)
        elif func in ["atan", "arctan"]:
            return r"\arctan%s" % self.parenthesis(args)
        elif func in ["arcsinh"]:
            return r"\sinh^{-1}%s" % self.parenthesis(args)
        elif func in ["arccosh"]:
            return r"\cosh^{-1}%s" % self.parenthesis(args)
        elif func in ["arctanh"]:
            return r"\tanh^{-1}%s" % self.parenthesis(args)
        elif func in ["power", "pow"]:
            args = [arg.strip() for arg in args.split(',')]
            if '+' in args[0] or '-' in args[0]:
                return self.power(self.parenthesis(args[0]), args[1])
            else:
                return self.power(args[0], args[1])
        elif func in ["divide"]:
            args = [arg.strip() for arg in args.split(',')]
            return self.division(args[0], args[1])
        elif func in ["abs", "fabs"]:
            return r"|%s|" % args
        elif func in ["exp"]:
            return r"e^{%s}" % args

        # Additionnal functions (convention names, not in numpy library)
        elif func in ["kronecher", "kron"]:
            return r"\delta_{%s}" % args

        # Integrals
        # TODO : add this integral in a visit_tripOp function???
        elif func in ["quad"]:
            (f, a, b) = list(map(self.visit, n.args))
            return r"\int_{%s}^{%s} %s%s d%s" % (
                a,
                b,
                f,
                self.parenthesis(self.dummy_var),
                self.dummy_var,
            )
        #
        # Sum
        elif func in ["sum"]:
            if blist:
                return "\sum_{%s=%s}^{%s} %s" % (
                    kwargs["iterator"],
                    kwargs["min"],
                    kwargs["max"],
                    kwargs["content"],
                )
            else:
                return r"\sum %s" % (args)

        # Recurrent operator names
        elif func in ["f", "g", "h"]:
            return r"%s{%s}" % (func, self.parenthesis(args))

        else:
            return self.operator(func, args)

    def visit_Name(self, n):
        """Special features:
        - Recognize underscripts in identifiers names (default: underscore)
        - Recognize upperscripts in identifiers names (default: ˆ, valid in Python3)
        Note that using ˆ is not recommanded in variable names because it may
        be confused with the operator ^, but in some special cases of extensively
        long formulas with lots of indices, it may help the readability of the
        code
        """

        u = n.id.count(self.upper)
        if u > 1:
            if self.verbose:
                warn("Only one upperscript character supported per identifier")

        def build_tree(expr, level=1):
            """Builds a tree out of a valid Python identifier, according to the
            following proposed formalism:

            Formalism
            ----------
                Python -> Real

                k_i_j  -> k_i,j
                k_i__j -> k_(i_j)
                k_iˆj -> k_i^j
                k_iˆˆj -> k_(i^j)
                k_i__1_i__2ˆj__1ˆˆj__2 -> k_(i_1,i_2)^(j_1,j_2)

            Even if one may agree that this last expression isn't a very
            readable variable name.

            The idea behind this is that I want my Python formula to be the
            same objects as the LaTeX formula I write in my reports / papers

            It allows me to:
            - gain time
            - check my Python formula (once printed LaTeX is much more readable
            that a multiline Python expression)

            """

            #            sep0 = '[{0}][{1}]'.format(self.lower,self.upper)
            sep = "[{0}{1}]".format(self.lower, self.upper)
            s = re.split(r"(?<!{0})({0}{{{1}}})(?!{0})".format(sep, level),
                         expr)  # Also returns the pattern n
            t = {}  # build tree
            if self.verbose:
                uprint("  " * (level - 1), "val:", self.convert_symbols(s[0]))
            t["val"] = self.convert_symbols(s[0])
            t["low"] = []
            t["up"] = []
            for i in range(1, len(s), 2):
                p = s[i]
                if p == self.lower * level:
                    if self.verbose:
                        uprint("  " * (level - 1), "low:", s[i + 1])
                    t["low"].append(build_tree(s[i + 1], level + 1))
                elif p == self.upper * level:
                    if self.verbose:
                        uprint("  " * (level - 1), "up:", s[i + 1])
                    t["up"].append(build_tree(s[i + 1], level + 1))
                else:
                    raise ValueError("Undetected separator")
            return t

        def read_tree(t):
            """ Write a LaTeX readable name """
            r = t["val"]
            if t["low"] != []:
                #                child = [self.group(read_tree(tc)) for tc in t['low']]
                child = [read_tree(tc) for tc in t["low"]]
                r += "_{0}".format(self.group(",".join(child)))
            if t["up"] != []:
                #                child = [self.group(read_tree(tc)) for tc in t['up']]
                child = [read_tree(tc) for tc in t["up"]]
                r += "^{0}".format(self.group(",".join(child)))
            return r

        return read_tree(build_tree(n.id))

    #    def convert_underscores(self, expr):
    #
    #        s = expr.split(self.lower)
    #
    #        for i, m in enumerate(s):
    #            s[i] = self.convert_symbols(m)
    #
    #        return s

    def convert_symbols(self, expr):
        m = expr
        # Standard greek letters
        if m in [
                "alpha",
                "beta",
                "gamma",
                "delta",
                "epsilon",
                "zeta",
                "eta",
                "theta",
                "iota",
                "kappa",
                "mu",
                "nu",
                "xi",
                "pi",
                "rho",
                "sigma",
                "tau",
                "phi",
                "chi",
                "psi",
                "omega",
                "Gamma",
                "Delta",
                "Theta",
                "Lambda",
                "Xi",
                "Pi",
                "Sigma",
                "Upsilon",
                "Phi",
                "Psi",
                "Omega",
        ]:
            m = r"\%s" % m

        elif m in ["eps"]:
            m = r"\epsilon"

        elif m in [
                "lbd"
        ]:  # lambda is not a valid identifier in Python so people use other things
            m = r"\lambda"

        elif m in ["Lbd"]:
            m = r"\Lambda"

        elif m in ["inf", "infinity", "infty"]:
            m = r"\infty"

        # Replace Delta even if not full word  - Allow for expressions such as
        # ΔE
        elif "Delta" in m:
            m = m.replace("Delta", "\Delta ")

        return m

    def visit_UnaryOp(self, n):
        # Note: removed space between {0} and {1}... so that 10**-3 yields
        # $$10^{-3}$$ and not $$10^{- 3}$$

        if self.prec(n.op) > self.prec(n.operand):
            return r"{0}{1}".format(self.visit(n.op),
                                    self.parenthesis(self.visit(n.operand)))
        else:
            return r"{0}{1}".format(self.visit(n.op), self.visit(n.operand))

    def prec_UnaryOp(self, n):
        return self.prec(n.op)

    def visit_BinOp(self, n):

        if self.prec(n.op) > self.prec(n.left):
            left = self.parenthesis(self.visit(n.left))
        elif isinstance(n.op, ast.Pow) and self.prec(n.op) == self.prec(
                n.left):
            # Special case for power, which needs parentheses when combined to the left
            left = self.parenthesis(self.visit(n.left))
        else:
            left = self.visit(n.left)
        if self.prec(n.op) > self.prec(n.right):
            right = self.parenthesis(self.visit(n.right))
        else:
            right = self.visit(n.right)

        # Special binary operators
        if isinstance(n.op, ast.Div):
            if self.simplify_fractions:
                left_is_int = self.looks_like_int(left)
                right_is_int = self.looks_like_int(right)
                if left_is_int or right_is_int:
                    if left_is_int and right_is_int:
                        return self.division("%d" % int(float(left)),
                                             "%d" % int(float(right)))
                    elif left_is_int:
                        return self.division("%d" % int(float(left)),
                                             self.visit(n.right))
                    else:
                        return self.division(self.visit(n.left),
                                             "%d" % int(float(right)))
            return self.division(self.visit(n.left), self.visit(n.right))
        elif isinstance(n.op, ast.FloorDiv):
            return r"\left\lfloor\frac{%s}{%s}\right\rfloor" % (
                self.visit(n.left),
                self.visit(n.right),
            )
        elif isinstance(n.op, ast.Pow):
            return self.power(left, self.visit(n.right))
        elif isinstance(n.op, ast.Mult):

            def looks_like_float(a):
                # Check if 'a' looks like a float
                # Detect: 'float', '{float}', 'int', '{int}'
                if a.startswith('{'):
                    a = a[1:].split('}')[0]
                try:
                    float(a)
                    return True
                except ValueError:
                    return False

            left_is_float = looks_like_float(left)
            right_is_float = looks_like_float(right)

            # Get multiplication operator. Force x if floats are involved
            if left_is_float or right_is_float:
                operator = self.tex_multiplier
            else:  # get standard Mult operator (see visit_Mult)
                operator = self.visit(n.op)

            if self.simplify_multipliers:

                # We simplify in some cases, for instance: a*2 -> 2a
                # First we need to know if both terms start with numbers
                if left[0] == "-" or left[0] == "{":
                    left_starts_with_digit = left[1].isdigit()
                else:
                    left_starts_with_digit = left[0].isdigit()
                if right[0] == "-" or right[0] == "{":
                    right_starts_with_digit = right[1].isdigit()
                else:
                    right_starts_with_digit = right[0].isdigit()

                # Simplify
                # ... simplify (a*2 --> 2a)
                if right_is_float and not left_starts_with_digit:
                    return r"{0}{1}".format(right, left)
                # ... simplify (2*a --> 2a)
                elif left_is_float and not right_starts_with_digit:
                    return r"{0}{1}".format(left, right)
                else:
                    return r"{0}{1}{2}".format(left, operator, right)
            else:
                return r"{0}{1}{2}".format(left, operator, right)
        else:
            return r"{0}{1}{2}".format(left, self.visit(n.op), right)

    def prec_BinOp(self, n):
        return self.prec(n.op)

    def visit_Sub(self, n):
        return "-"

    def visit_Add(self, n):
        return "+"

    def visit_Mult(self, n):
        return r" "

    def visit_Mod(self, n):
        return "\\bmod"

    def visit_LShift(self, n):
        return self.operator("shiftLeft")

    def visit_RShift(self, n):
        return self.operator("shiftRight")

    def visit_BitOr(self, n):
        return self.operator("or")

    def visit_BitXor(self, n):
        return self.operator("xor")

    def visit_BitAnd(self, n):
        return self.operator("and")

    def visit_Invert(self, n):
        return self.operator("invert")

    def visit_Not(self, n):
        return "\\neg "

    def visit_UAdd(self, n):
        return "+"

    def visit_USub(self, n):
        return "-"

    def visit_Num(self, n):
        if self.simplify_fractions:
            if any([n.n == key for key in fracs.keys()]):
                return r"{0}\frac{{{1}}}{{{2}}}".format(*fracs[n.n])
        if self.looks_like_int(n.n):
            return "%d" % n.n
        return str(n.n)

    # New visits
    def visit_Assign(self, n):
        " Rewrite Assign function (instead of executing it)"
        return r"%s=%s" % (self.visit(n.targets[0]), self.visit(n.value))

    def visit_Compare(self, n):
        " Rewrite Compare function (instead of executing it)"

        def visit_Op(op):
            " Note : not called by visit like other visit functions"
            if isinstance(op, ast.Lt):
                return "<"
            elif isinstance(op, ast.LtE):
                return "<="
            elif isinstance(op, ast.Gt):
                return ">"
            elif isinstance(op, ast.GtE):
                return ">="
            elif isinstance(op, ast.Eq):
                return "="
            else:
                raise ValueError("Unknown comparator", op.__class__)

        return r"%s%s" % (
            self.visit(n.left),
            "".join([
                "%s%s" % (visit_Op(n.ops[i]), self.visit(n.comparators[i]))
                for i in range(len(n.comparators))
            ]),
        )

    # Default
    def generic_visit(self, n):
        if isinstance(n, ast.AST):
            return r"" % (
                n.__class__.__name__,
                ", ".join(map(self.visit, [getattr(n, f) for f in n._fields])),
            )
        else:
            return str(n)

    def generic_prec(self, n):
        return 0

    # LaTeX blocs
    def brackets(self, expr):
        """ Enclose expr in {...} """
        return r"{{{0}}}".format(expr)

    def group(self, expr):
        """ Returns expr, add brackets if needed"""
        if len(expr) == 1:
            return expr
        else:
            return self.brackets(expr)

    def parenthesis(self, expr):
        return r"\left({0}\right)".format(expr)

    def power(self, expr, power):
        return r"{0}^{1}".format(self.group(expr), self.group(power))

    def division(self, up, down):
        return r"\frac{0}{1}".format(self.brackets(up), self.brackets(down))

    def sqrt(self, args):
        return r"\sqrt{0}".format(self.brackets(args))

    def operator(self, func, args=None):
        if args is None:
            return r"\operatorname{{{0}}}".format(func)
        else:
            return r"\operatorname{{{0}}}{1}".format(func,
                                                     self.parenthesis(args))


def preprocessing(expr, simplify):
    """Pre-process a string."""

    # replace unicode values (so that even a Python 2 pytexit can parse formula
    #  with unicode, valid in Python 3 only)
    for u in unicode_tbl:
        expr = expr.replace(u, unicode_tbl[u])

    # remove unnessary calls to libraries
    # Example: np.exp(-3) would be read exp(-3)
    expr = expr.strip()  # remove spaces on the side
    for m in clear_modules:
        # Todo: some regexp here. re(<(+- */)). To make sure we're not removing
        # a variable name
        expr = expr.replace(m + ".", "")

    return expr


def replace_scientific(s):
    """ Replace 'NUMBER e NUMBER' with powers of 10 """

    regexp = re.compile(r"(\d*\.{0,1}\d+)[eE]([-+]?\d*\.{0,1}\d+)")

    matches = regexp.findall(s)
    splits = regexp.split(s)
    assert len(splits) == (len(matches) + 1) + (2 * len(matches))
    #                     splitted groups             prefactor, exponent

    new_s = ""
    # loop over all match and replace
    # ... I didnt find any better way to do that given that we want a conditional
    # ... replace (.sub wouldnt work)
    for i, (prefactor, exponent) in enumerate(matches):
        new_s += splits[3 * i]

        if float(prefactor) == 1.0:
            new_s += r"10**{}".format(exponent)
        else:
            new_s += r"{}*10**{}".format(prefactor, exponent)
    if len(splits) % 3 == 1:  # add last ones
        new_s += splits[-1]
    return new_s


def simplify(s):
    """ Cleans the generated text in post-processing """

    # Remove unecessary parenthesis?
    # -------------
    # TODO: look for groups that looks like \(\([\(.+\)]*\)\ ),
    # (2 pairs of external parenthesis around any number (even 0) of closed pairs
    # of parenthesis)  -> then remove one of the the two external parenthesis
    # TRied with re.findall(r'\(\([^\(^\)]*(\([^\(^\)]+\))*[^\(^\)]*\)\)', s)  but
    # it doesnt work. One should better try to look for inner pairs and remove that
    # one after one..

    # Replace '\left(NUMBER\right)' with 'NUMBER'
    # ------------
    s = re.sub(r"\\left\(([\d\.]+)\\right\)", r"\1", s)

    return s



import os
import re
import shutil
import stat
import sys
import tempfile
import warnings
from os.path import dirname, abspath, join, basename, splitext
from rich.console import Console
from rich.markdown import Markdown
try:
    from woc.decorator import decorator
    from woc.config import Config
except:
    from .config import Config
    from .decorator import decorator
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional

import requests

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

_canonicalize_regex = re.compile("[-_]+")

SHELL = os.getenv("SHELL", "")
MACOS = sys.platform == "darwin"
WINDOWS = sys.platform.startswith("win") or (sys.platform == "cli"
                                             and os.name == "nt")


def redirect(hide: bool = False):
    """
    use `2>&1` to redirect the output of the error message to standard output

    :param hide:
    :return:
    """
    if WINDOWS:
        return ''
    elif hide:
        # the final & is to make the program continue when terminal exit
        return ' 2>&1 &'
    else:
        return ''


def get_pure_filename(fpath: str):
    """Get pure filename

    :param fpath: woc/woc/woc.txt
    :return: woc
    """
    return splitext(basename(fpath))[0]


def canonicalize_name(name: str) -> str:
    return _canonicalize_regex.sub("-", name).lower()


def module_name(name: str) -> str:
    return canonicalize_name(name).replace(".", "_").replace("-", "_")


def _del_ro(action: Callable, name: str, exc: Exception) -> None:
    os.chmod(name, stat.S_IWRITE)
    os.remove(name)


@contextmanager
def temporary_directory(*args: Any, **kwargs: Any) -> Iterator[str]:
    name = tempfile.mkdtemp(*args, **kwargs)

    yield name

    shutil.rmtree(name, onerror=_del_ro)


def get_cert(config: Config, repository_name: str) -> Optional[Path]:
    cert = config.get(f"certificates.{repository_name}.cert")
    if cert:
        return Path(cert)
    else:
        return None


def get_client_cert(config: Config, repository_name: str) -> Optional[Path]:
    client_cert = config.get(f"certificates.{repository_name}.client-cert")
    if client_cert:
        return Path(client_cert)
    else:
        return None


def _on_rm_error(func: Callable, path: str, exc_info: Exception) -> None:
    if not os.path.exists(path):
        return

    os.chmod(path, stat.S_IWRITE)
    func(path)


def safe_rmtree(path: str) -> None:
    if Path(path).is_symlink():
        return os.unlink(str(path))

    shutil.rmtree(path, onerror=_on_rm_error)


def merge_dicts(d1: Dict, d2: Dict) -> None:
    for k, v in d2.items():
        if k in d1 and isinstance(d1[k], dict) and isinstance(d2[k], Mapping):
            merge_dicts(d1[k], d2[k])
        else:
            d1[k] = d2[k]


def download_file(
    url: str,
    dest: str,
    session: Optional[requests.Session] = None,
    chunk_size: int = 1024,
) -> None:
    get = requests.get if not session else session.get

    with get(url, stream=True) as response:
        response.raise_for_status()

        with open(dest, "wb") as f:
            for chunk in response.iter_content(chunk_size=chunk_size):
                if chunk:
                    f.write(chunk)


def paths_csv(paths: List[Path]) -> str:
    return ", ".join('"{}"'.format(str(c)) for c in paths)


def is_dir_writable(path: Path, create: bool = False) -> bool:
    try:
        if not path.exists():
            if not create:
                return False
            path.mkdir(parents=True, exist_ok=True)

        with tempfile.TemporaryFile(dir=str(path)):
            pass
    except OSError:
        return False
    else:
        return True


def render_markdown(file):
    console = Console()
    with open(file, 'r') as readme:
        markdown = Markdown(readme.read())
    console.print(markdown)


@decorator
def deprecated(func, *args, **kwargs):
    """ Marks a function as deprecated. """
    warnings.warn(
        f"{func} is deprecated and should no longer be used.",
        DeprecationWarning,
        stacklevel=3,
    )
    return func(*args, **kwargs)


def deprecated_option(option_name, message=""):
    """ Marks an option as deprecated. """
    def caller(func, *args, **kwargs):
        if option_name in kwargs:
            warnings.warn(
                f"{option_name} is deprecated. {message}",
                DeprecationWarning,
                stacklevel=3,
            )

        return func(*args, **kwargs)

    return decorator(caller)



from typing import Any
from typing import Dict

from .config_source import ConfigSource


class DictConfigSource(ConfigSource):
    def __init__(self) -> None:
        self._config = {}

    @property
    def config(self) -> Dict[str, Any]:
        return self._config

    def add_property(self, key: str, value: Any) -> None:
        keys = key.split(".")
        config = self._config

        for i, key in enumerate(keys):
            if key not in config and i < len(keys) - 1:
                config[key] = {}

            if i == len(keys) - 1:
                config[key] = value
                break

            config = config[key]

    def remove_property(self, key: str) -> None:
        keys = key.split(".")

        config = self._config
        for i, key in enumerate(keys):
            if key not in config:
                return

            if i == len(keys) - 1:
                del config[key]

                break

            config = config[key]



import sys

try:
    from importlib import metadata
except ImportError:
    # compatibility for python <3.8
    import importlib_metadata as metadata  # noqa

WINDOWS = sys.platform == "win32"


def decode(string, encodings=None):
    if not isinstance(string, bytes):
        return string

    encodings = encodings or ["utf-8", "latin1", "ascii"]

    for encoding in encodings:
        try:
            return string.decode(encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            pass

    return string.decode(encodings[0], errors="ignore")


def encode(string, encodings=None):
    if isinstance(string, bytes):
        return string

    encodings = encodings or ["utf-8", "latin1", "ascii"]

    for encoding in encodings:
        try:
            return string.encode(encoding)
        except (UnicodeEncodeError, UnicodeDecodeError):
            pass

    return string.encode(encodings[0], errors="ignore")


def to_str(string):
    return decode(string)


def list_to_shell_command(cmd):
    return " ".join(
        f'"{token}"' if " " in token and token[0] not in {"'", '"'} else token
        for token in cmd)



import os

from pathlib import Path

try:
    from woc.appdirs import user_cache_dir
except:
    from woc.appdirs import user_cache_dir
from woc.appdirs import user_config_dir
from .appdirs import user_data_dir

CACHE_DIR = user_cache_dir("woc")
DATA_DIR = user_data_dir("woc")
CONFIG_DIR = user_config_dir("woc")

REPOSITORY_CACHE_DIR = Path(CACHE_DIR) / "cache" / "repositories"


def data_dir() -> Path:
    if os.getenv("WOC_HOME"):
        return Path(os.getenv("WOC_HOME")).expanduser()

    return Path(user_data_dir("woc", roaming=True))



#!/usr/bin/python3
# *_*coding:utf-8 *_*
import os
import sys
import subprocess
import traceback
from datetime import datetime
from pathlib import Path
from tqdm import tqdm
from rich.progress import track
from rich.console import Console
from os.path import dirname, abspath, join, basename, splitext

from woc.downloader import downloading
try:
    from woc.math2latex import py2tex
    from woc.helpers import render_markdown, redirect, get_pure_filename
except:
    from .math2latex import py2tex
    from .helpers import render_markdown
import webbrowser

ROOT = dirname(abspath(__file__))
import click
from click_help_colors import HelpColorsGroup, HelpColorsCommand

# CONSTANTS
CLICK_CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])
CLICK_CONTEXT_SETTINGS_NO_HELP = dict(help_option_names=[])


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    package_conf = {}
    with open(join(ROOT, "__version__.py")) as f:
        exec(f.read(), package_conf)
    # click.secho(package_conf['__version__'], fg='green')
    click.secho(package_conf['__version__'], blink=True, bold=True)
    ctx.exit()


@click.group(chain=True,
             cls=HelpColorsGroup,
             context_settings=CLICK_CONTEXT_SETTINGS,
             help_headers_color='yellow',
             help_options_color='magenta',
             help_options_custom_colors={
                 'install': 'red',
                 'pip': 'cyan',
                 'pipenv': 'red',
                 'git': 'red',
                 'pypi': 'green',
                 'run': 'magenta',
                 'hexo': 'green',
                 'tree': 'cyan',
                 'docs': 'cyan',
                 'kit': 'magenta',
             })
@click.option('-v',
              '--version',
              is_flag=True,
              callback=print_version,
              expose_value=False,
              is_eager=True,
              help='Show version')
def cli():
    """\b
██╗    ██╗ ██████╗  ██████╗
██║    ██║██╔═══██╗██╔════╝
██║ █╗ ██║██║   ██║██║
██║███╗██║██║   ██║██║
╚███╔███╔╝╚██████╔╝╚██████╗
 ╚══╝╚══╝  ╚═════╝  ╚═════╝
    """


# https://manytools.org/hacker-tools/ascii-banner/


@cli.command(context_settings=dict(ignore_unknown_options=True, ),
             cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='pipenv virtual environment pipeline')
@click.option('-c',
              '--create',
              is_flag=True,
              help="create pipenv environment base current directory")
@click.option('-d',
              '--delete',
              is_flag=True,
              help="delete pipenv environment base current directory")
def pipenv(create, delete):
    """Examples:

        \b
                create pipenv environment base current directory:
                    - woc pipenv -c | --create
                delete pipenv environment base current directory:
                    - woc pipenv -d | --delete
        """
    CHECK_PIPENV = os.system('pipenv --venv')
    if CHECK_PIPENV == 0:
        # exist pipenv environment
        pass
    if create:
        FILE = join(ROOT, 'scripts', 'create.sh')
        subprocess.run(f'bash {FILE}'.split())
    if delete:
        FILE = join(ROOT, 'scripts', 'delete.sh')
        subprocess.run(f'bash {FILE}'.split())
    if not (create or delete):
        click.secho(
            "I don't know what you're trying to do. Do you know what you're doing...",
            fg='red')


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             help='render a beautiful tree with given path')
@click.argument('path', nargs=-1)
def tree(path):
    '''render a beautiful tree with given path'''
    FILE = join(ROOT, 'tree.py')
    DIR = path[0] if path else os.getcwd()
    subprocess.run(f'python3 {FILE} {DIR}'.split())


@cli.command(context_settings=dict(ignore_unknown_options=True, ),
             cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='install python package')
@click.argument('pkgs', nargs=-1, required=True)
@click.option('-a',
              '--ali',
              is_flag=True,
              show_default=True,
              help="use aliyun pypi source")
@click.option('-d',
              '--douban',
              is_flag=True,
              show_default=True,
              help="use douban pypi source")
@click.option('-H',
              '--hide',
              is_flag=True,
              show_default=True,
              help="hide message on terminal")
@click.option('-v',
              '--virtualenv',
              is_flag=True,
              show_default=True,
              help="install package in pipenv virtualenv")
@click.option('-u',
              '--upgrade',
              is_flag=True,
              show_default=True,
              help="upgrade package")
def pip(pkgs, ali, douban, upgrade, hide, virtualenv):
    if upgrade:
        UPGRADE = '--upgrade'
    else:
        UPGRADE = ''

    if pkgs[0] in ['requirements.txt', 'requirements-dev.txt']:
        file = pkgs[0]
        with open(file, 'r') as f:
            pkgs = [pkg.strip() for pkg in f.readlines() if len(pkg) > 0]
    pkgs = [pkg for pkg in pkgs if not str(pkg).__contains__('-')]

    REDIRECT_SEG = redirect(hide)

    if virtualenv:
        # pipenv path
        PIPENV = subprocess.getoutput('pipenv --venv')
        PIP = PIPENV + '/bin/pip'
    else:
        PIP = 'pip3'

    # trying to upgrade pip first
    os.system(
        f'{PIP} install pip --upgrade -i https://mirrors.aliyun.com/pypi/simple'
        f' {REDIRECT_SEG}')

    # for pkg in tqdm(pkgs):
    for pkg in track(pkgs, description=''):
        if ali:
            os.system(f'{PIP} install {pkg} {UPGRADE} -i '
                      f'https://mirrors.aliyun.com/pypi/simple {REDIRECT_SEG}')
        elif douban:
            os.system(
                f'{PIP} install {pkg} {UPGRADE} -i https://pypi.douban.com/simple'
                f' {REDIRECT_SEG}')
        else:
            os.system(
                f'{PIP} install {pkg} {UPGRADE} -i https://pypi.org/simple {REDIRECT_SEG}'
            )


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='simplified git pipeline')
@click.option('-p', '--push', is_flag=True, help='push change to remote')
@click.option('-m', '--msg', help='message')
@click.option('-doc', '--doc', is_flag=True, help='show git tutorials')
@click.option('-c', '--cache', is_flag=True, help='remove cached files')
@click.option('-l', '--lock', is_flag=True, help='remove index.lock')
@click.option('-t', '--tag', help='git tag')
def git(push, msg, cache, lock, tag, doc):

    if push:
        if not msg:
            # If no massage is given, to use the current time instead
            msg = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        os.system(f'yapf -irp .;git add . --all;git commit -m "'
                  f'{msg}";git push')

    if lock:
        # fix bug: fatal: Unable to create 'xxx/.git/index.lock': File exists.
        subprocess.run('rm -f ./.git/index.lock'.split())
    if cache:
        subprocess.run('git rm -r --cache .'.split())

    if tag:
        COMMAND = f'''git tag --annotate "{tag}" --message "{msg if msg else tag}"'''
        os.system(COMMAND)

    if doc:
        render_markdown(join(ROOT, 'resources', 'GitTutorials.md'))


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='hexo pipeline')
@click.option('-d', '--deploy', is_flag=True, help='deploy hexo blog')
def hexo(deploy):
    if deploy:
        FILE = join(ROOT, 'scripts', 'deploy.sh')
        with Console().status("[bold green]"):
            subprocess.run(f'bash {FILE}'.split())
    else:
        click.secho(
            "I don't know what you're trying to do. Do you know what you're doing...",
            fg='red')


def render_alias():
    render_markdown(join(ROOT, 'resources', 'Alias.md'))


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='show document')
@click.option('-kr', '--keras', is_flag=True, help="show keras doc")
@click.option('-sl', '--sklearn', is_flag=True, help="show sklearn doc")
@click.option('-pd', '--pandas', is_flag=True, help="show pandas doc")
@click.option('-tf', '--tensorflow', is_flag=True, help="show tensorflow doc")
@click.option('-tc', '--torch', is_flag=True, help="show pytorch doc")
@click.option('-md', '--markdown', is_flag=True, help="show markdown doc")
@click.option('-py', '--python', is_flag=True, help="show python doc")
@click.option('-pl', '--plotly', is_flag=True, help="show plotly doc")
@click.option('-ple',
              '--plotly_express',
              is_flag=True,
              help="show plotly "
              "express doc")
def docs(markdown, torch, tensorflow, pandas, sklearn, keras, python, plotly,
         plotly_express):
    """👀Documents👀"""
    if markdown:
        webbrowser.open('https://www.songzhijun.com/posts/89757140/')
    if pandas:
        webbrowser.open('https://pandas.pydata.org/docs/user_guide/')
    if tensorflow:
        webbrowser.open('https://www.tensorflow.org/addons/api_docs/python/')
    if torch:
        webbrowser.open('https://pytorch.org/tutorials/')
    if keras:
        webbrowser.open('https://keras.io/examples/')
    if sklearn:
        webbrowser.open('https://scikit-learn.org/stable/')
    if python:
        webbrowser.open('https://docs.python.org/3/')
    if plotly:
        webbrowser.open('https://plotly.com/python/plotly-fundamentals/')
    if plotly_express:
        webbrowser.open('https://plotly.com/python/plotly-express/#')

    # webbrowser.open('https://www.songzhijun.com')


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='install something useful')
@click.argument('pkg', nargs=1, required=True)
def install(pkg):
    """Supports:

    \b
            node | apt | brew
    """
    if pkg == 'node':
        if sys.platform.startswith('darwin'):
            with Console().status("[bold green]"):
                subprocess.run(f'brew install node'.split())
        elif sys.platform.startswith('win'):
            pass
        else:
            # https://developer.aliyun.com/article/760687
            # 先安装node包管理器nvm
            with Console().status("[bold green]install nvm..."):
                subprocess.run(
                    f'wget -qO- https://raw.githubusercontent.com/nvm-sh/nvm/v0.38.0/install.sh | bash'
                    .split())
                subprocess.run(f'export NVM_DIR="$HOME/.nvm"'.split())
            subprocess.run(f'nvm --version'.split())
            # 安装最新版node
            with Console().status("[bold green]install node..."):
                subprocess.run(f'nvm install node'.split())
    elif pkg == 'brew':
        os.system(
            '/bin/bash -c "$(curl -fsSL '
            'https://raw.githubusercontent.com/Homebrew/install/HEAD/install'
            '.sh)"')
    elif pkg == 'apt':
        FILE = join(ROOT, 'scripts', 'aptinstall.sh')
        with Console().status("[bold green]installing..."):
            subprocess.run(f'bash {FILE}'.split())
    else:
        click.secho(
            "I don't know what you're trying to do. Do you know what you're doing...",
            fg='red')


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='toolkits...')
@click.option('-ip', '--ip', is_flag=True, help='checkout ip address')
@click.option('-serv',
              '--server',
              is_flag=True,
              help='server directory to browser')
def kit(ip, server):
    if ip:
        os.system('ifconfig | grep "inet " | grep -v 127.0.0.1')
    if server:
        os.system('python3 -m http.server --directory ./')


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='config some environment')
@click.option('-jp',
              '--jupyter',
              is_flag=True,
              help='config jupyter notebook extensions and theme')
@click.option('-al', '--alias', is_flag=True, help='print frequent used alias')
def config(jupyter, alias):

    if jupyter:
        FILE = join(ROOT, 'scripts', 'setnotebook.sh')
        with Console().status("[bold green]configing jupyter..."):
            subprocess.run(f'bash {FILE}'.split())
    if alias:
        render_alias()


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='manage pypi package')
@click.option('-p', '--publish', is_flag=True, help='publish package to pypi')
@click.option('-c', '--clean', is_flag=True, help='clean compile files')
def pypi(publish, clean):
    """Examples:

    \b
            publish package to pypi:
                - woc pypi -p | --publish
            clean pypi build output:
                - woc pypi -c | --clean
            combine activity chain:
                - woc pypi -pc
    """

    if publish:
        FILE = join(ROOT, 'scripts', 'publish.sh')
        subprocess.run(f'bash {FILE}'.split())

    if clean:
        FILE = join(ROOT, 'scripts', 'clean.sh')
        subprocess.run(f'bash {FILE}'.split())

    if not (publish or clean):
        click.secho(
            "I don't know what you're trying to do. Do you know what you're doing...",
            fg='red')


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             help='run python script')
@click.argument('script', nargs=1, required=True)
@click.option('-d', '--logdir', default='logs', help='log directory')
@click.option('-b',
              '--background',
              is_flag=True,
              show_default=True,
              help="run program in background")
@click.option('-s',
              '--show',
              is_flag=True,
              show_default=True,
              help="show output in terminal")
def run(script, logdir, background, show):
    # pipenv path
    PIPENV = subprocess.getoutput('which pipenv')
    CHECK_PIPENV = os.system('pipenv --venv')
    LOG_PATH = f'''{logdir}/{get_pure_filename(script)}'''
    LOG_FILE = f'''{LOG_PATH}/{datetime.now().strftime(
        "%Y-%m-%d:%H:%M:%S")}.log'''
    Path(LOG_PATH).mkdir(parents=True, exist_ok=True)

    REDIRECT_SEG = redirect(background)

    if CHECK_PIPENV == 0:
        # if exist pipenv environment
        if show:
            os.system(f'''{PIPENV} run python3 {script}''')
        else:
            os.system(
                f'''{PIPENV} run python3 {script} >>{LOG_FILE} {REDIRECT_SEG}'''
            )
    else:
        if show:
            os.system(f'''python3 {script}''')
        else:
            os.system(f'''python3 {script} >>{LOG_FILE} {REDIRECT_SEG}''')


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             help='download url')
@click.argument('args', nargs=-1, required=True)
@click.option('-d',
              '--dir',
              type=click.Path(),
              default="./",
              show_default=True,
              help="download directory")
def download(args, dir):
    downloading(args, dir)


@cli.command(cls=HelpColorsCommand,
             help_options_color='cyan',
             short_help='render math expression into latex')
@click.argument('math_expression', nargs=1, required=True)
def latex(math_expression):
    '''Example:

    \b
    woc latex 'x = 2*sqrt(2*pi*k*T_e/m_e)*(DeltaE/(k*T_e))**2*a_0**2'
    '''
    latex = py2tex(math_expression)
    # latex = py2tex('x = 2*sqrt(2*pi*k*T_e/m_e)*(DeltaE/(k*T_e))**2*a_0**2')
    Console().print(latex, style='green')


def execute():
    try:
        cli()
    except Exception as error:
        traceback.print_exc()


if __name__ == "__main__":
    execute()



import os
import re

from copy import deepcopy
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

try:
    from woc.locations import CACHE_DIR
except:
    from locations import CACHE_DIR

from .config_source import ConfigSource
from .dict_config_source import DictConfigSource

_NOT_SET = object()


def boolean_validator(val: str) -> bool:
    return val in {"true", "false", "1", "0"}


def boolean_normalizer(val: str) -> bool:
    return val in ["true", "1"]


class Config:

    default_config = {
        "cache-dir": str(CACHE_DIR),
        "virtualenvs": {
            "create": True,
            "in-project": None,
            "path": os.path.join("{cache-dir}", "virtualenvs"),
            "options": {
                "always-copy": False,
                "system-site-packages": False
            },
        },
        "experimental": {
            "new-installer": True
        },
        "installer": {
            "parallel": True
        },
    }

    def __init__(self,
                 use_environment: bool = True,
                 base_dir: Optional[Path] = None) -> None:
        self._config = deepcopy(self.default_config)
        self._use_environment = use_environment
        self._base_dir = base_dir
        self._config_source = DictConfigSource()
        self._auth_config_source = DictConfigSource()

    @property
    def name(self) -> str:
        return str(self._file.path)

    @property
    def config(self) -> Dict:
        return self._config

    @property
    def config_source(self) -> ConfigSource:
        return self._config_source

    @property
    def auth_config_source(self) -> ConfigSource:
        return self._auth_config_source

    def set_config_source(self, config_source: ConfigSource) -> "Config":
        self._config_source = config_source

        return self

    def set_auth_config_source(self, config_source: ConfigSource) -> "Config":
        self._auth_config_source = config_source

        return self

    def merge(self, config: Dict[str, Any]) -> None:
        from woc.utils.helpers import merge_dicts

        merge_dicts(self._config, config)

    def all(self) -> Dict[str, Any]:
        def _all(config: Dict, parent_key: str = "") -> Dict:
            all_ = {}

            for key in config:
                value = self.get(parent_key + key)
                if isinstance(value, dict):
                    if parent_key != "":
                        current_parent = parent_key + key + "."
                    else:
                        current_parent = key + "."
                    all_[key] = _all(config[key], parent_key=current_parent)
                    continue

                all_[key] = value

            return all_

        return _all(self.config)

    def raw(self) -> Dict[str, Any]:
        return self._config

    def get(self, setting_name: str, default: Any = None) -> Any:
        """
        Retrieve a setting value.
        """
        keys = setting_name.split(".")

        # Looking in the environment if the setting
        # is set via a WOC_* environment variable
        if self._use_environment:
            env = "WOC_{}".format("_".join(k.upper().replace("-", "_")
                                           for k in keys))
            value = os.getenv(env, _NOT_SET)
            if value is not _NOT_SET:
                return self.process(self._get_normalizer(setting_name)(value))

        value = self._config
        for key in keys:
            if key not in value:
                return self.process(default)

            value = value[key]

        return self.process(value)

    def process(self, value: Any) -> Any:
        if not isinstance(value, str):
            return value

        return re.sub(r"{(.+?)}", lambda m: self.get(m.group(1)), value)

    def _get_normalizer(self, name: str) -> Callable:
        if name in {
                "virtualenvs.create",
                "virtualenvs.in-project",
                "virtualenvs.options.always-copy",
                "virtualenvs.options.system-site-packages",
                "installer.parallel",
        }:
            return boolean_normalizer

        if name == "virtualenvs.path":
            return lambda val: str(Path(val))

        return lambda val: val



# -*- coding: utf-8 -*-
"""

Convert a Python expression in a LaTeX formula

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import ast
import sys
import six

try:
    from woc.latexcore import LatexVisitor, preprocessing, simplify, \
        uprint, replace_scientific
    from woc.latexdocx import WordVisitor
    from woc.latexfortran import for2py
except:  # if run locally as a script
    from .latexcore import LatexVisitor, preprocessing, simplify, \
        uprint, replace_scientific
    from .latexdocx import WordVisitor
    from .latexfortran import for2py
try:
    import IPython.display
except:
    pass


def py2tex(
    expr,
    print_latex=True,
    print_formula=False,
    dummy_var="u",
    output="tex",
    tex_enclosure="$$",
    tex_multiplier=r"\times",
    simplify_output=True,
    upperscript="ˆ",
    lowerscript="_",
    verbose=False,
    simplify_fractions=False,
    simplify_ints=True,
    simplify_multipliers=True,
):
    """Return the LaTeX expression of a Python formula

    Parameters
    ----------
    expr: string
        a Python expression

    print_latex: boolean
        if True, prints the latex expression in the console

    dummy_var: string
        dummy variable displayed in integrals

    output: 'tex' / 'word'
        if 'tex', output latex formula. If word, output a Word MathTex formula
        (may be a little different)

    tex_enclosure: string
        enclosure for latex formula.

        Default: "$$""

    tex_multiplier: raw string
        multiplication operator for latex formula.

        r'\times':   2*2 -> 2 x 2 (Default)
        r'{\times}': 2*2 -> 2x2
        r'\cdot':    2*2 -> 2 · 2
        r'{\cdot}':  2*2 -> 2·2

    Other Parameters
    ----------------

    simplify_output: boolean
        if ``True``, simplify output. Ex::

            1x10^-5 --> 10^-5

        See :func:`~pytexit.core.core.simplify` for more information.
        Default ``True``

    simplify_ints: boolean
        if ``True``, simplify integers (useful for Python 2 expressions). Ex::

            1. --> 1

        See :class:`~pytexit.core.core.LatexVisitor` for more information.
        Default ``True``

    simplify_fractions: boolean
        if ``True``, simplify common fractions.  Ex::

            0.5 --> 1/2

        See :class:`~pytexit.core.core.LatexVisitor` for more information.
        Default ``False``

    simplify_multipliers: boolean
        if ``True``, simplify float multipliers during parsing. Ex::

            2*a  -> 2a

        See :class:`~pytexit.core.core.LatexVisitor` for more information.
        Default ``True``


    Returns
    -------

    returns the latex expression in raw text, to be used in your reports or
    to display in an IPython notebook

    Notes
    -----

    Will return ``'\\\\'`` instead of ``'\\'`` because we don't want those to be
    interpreted as regular expressions. Use ``print(result)`` to get the correct
    LaTex formula.

    See Also
    --------

    :func:`~pytexit.pytexit.for2tex`

    References
    ----------

    Initial work from Geoff Reedy on StackOverflow: https://stackoverflow.com/a/3874621/5622825 . Kudos.

    Similar projects:

    - https://github.com/iogf/lax  : "A pythonic way of writting latex."
    - https://github.com/JelteF/PyLaTeX : "A Python library for creating LaTeX files"
    - sympy can also write LaTeX output.

    """
    # Check inputs
    try:
        if sys.version_info > (3, ):
            assert isinstance(expr, str)
        else:
            assert isinstance(expr, (str, six.text_type))
    except AssertionError:
        raise ValueError("Input must be a string")

    expr = preprocessing(
        expr, simplify_output)  # removes unicode, module calls, etc.

    # replace scientific notation with power of 10 (this needs to be done in
    # preprocessing since the ast parser will replace 1e3 with 1000.0)
    if simplify_output:
        expr = replace_scientific(expr)

    # Parse
    pt = ast.parse(expr)
    if output == "tex":  # LaTex output
        Visitor = LatexVisitor(
            dummy_var=dummy_var,
            upperscript=upperscript,
            lowerscript=lowerscript,
            verbose=verbose,
            simplify_multipliers=simplify_multipliers,
            simplify_fractions=simplify_fractions,
            simplify_ints=simplify_ints,
            tex_multiplier=tex_multiplier,
        )
    elif output == "word":  # Word output
        Visitor = WordVisitor(
            dummy_var=dummy_var,
            upperscript=upperscript,
            lowerscript=lowerscript,
            verbose=verbose,
            simplify_multipliers=simplify_multipliers,
            simplify_fractions=simplify_fractions,
            simplify_ints=simplify_ints,
            tex_multiplier=tex_multiplier,
        )
    else:
        raise ValueError("Unexpected output: {0}".format(output))
    if isinstance(pt.body[0], ast.Expr):
        # To deal with cases such as 'x=something'
        # TODO : one single command to start the visit?
        s = Visitor.visit(pt.body[0].value)
    else:  # For Compare / Assign expressions
        s = Visitor.visit(pt.body[0])

    # Simplify if asked for
    if simplify_output:
        s = simplify(s)

    if output == "tex":
        s = tex_enclosure + s + tex_enclosure

    if print_formula:
        uprint(s)
    return s


def for2tex(a, **kwargs):
    """Converts FORTRAN formula to Python Formula

    Parameters
    ----------

    a: str
        FORTRAN formula

    Other Parameters
    ----------------

    kwargs: dict
        forwarded to :func:`~pytexit.py2tex` function. See :func:`~pytexit.py2tex`
        doc.

    Examples
    --------

    convert FORTRAN formula to LaTeX with for2tex::

        for2tex(r'2.8d-11 * exp(-(26500 - 0.5 * 1.97 * 11600 )/Tgas)')

    See Also
    --------

    :func:`~pytexit.core.fortran.for2py`, :func:`~pytexit.pytexit.py2tex`

    """

    return py2tex(for2py(a), **kwargs)


if __name__ == "__main__":

    ll = py2tex('x = 2*sqrt(2*pi*k*T_e/m_e)*(DeltaE/(k*T_e))**2*a_0**2')
    print(ll)



import dataclasses

from typing import Dict
from typing import Union


@dataclasses.dataclass(order=True, eq=True)
class Source:
    name: str
    url: str
    default: bool = dataclasses.field(default=False)
    secondary: bool = dataclasses.field(default=False)

    def to_dict(self) -> Dict[str, Union[str, bool]]:
        return dataclasses.asdict(self)



# -*- coding: utf-8 -*-
"""
Tools to deal with Word equations format


See Also
--------

if working with Word and Python, you may be interested in the python-docx
package although equations are not yet supported.

"""

from __future__ import absolute_import, division, print_function, unicode_literals

from .latexcore import LatexVisitor


class WordVisitor(LatexVisitor):
    """A variant of the LatexVisitor to create Word readable equations
    (to be inserted in Word equation tool)

    See Also
    --------

    if working with Word and Python, you may be interested in the python-docx
    package although equations are not yet supported.

    """

    # Word-readable blocks
    def group(self, expr):
        "Word will convert unnecessary parenthesis in equivalent LaTeX {} groups"
        return self.parenthesis(expr)

    def parenthesis(self, expr):
        "No spacing"
        return "({0})".format(expr)

    #    def power(self, expr, power):
    #        'no { }'
    #        return r'{0}^{1}'.format(expr, power)

    def division(self, up, down):
        "no frac"
        return r"({0}/{1})".format(up, down)

    def visit_Mult(self, n):
        "No spacing"
        return r"\cdot"

    #        return r''

    def sqrt(self, args):
        return r"\sqrt({0})".format(args)

    def operator(self, func, args=None):
        if args is None:
            return r"{0}".format(func)
        else:
            return r"{0}({1})".format(func, args)
